const {
  time,
  loadFixture,
  // setPrevRandao,
} = require('@nomicfoundation/hardhat-toolbox/network-helpers');
const { expect } = require('chai');
const { ethers } = require('hardhat');

describe('Work1', () => {
  const WORK1_PRICE = 100;
  const STAKE_DURATION = 60;
  const STAKE_AMOUNT = 50;
  async function deploy() {
    // Contracts are deployed using the first signer/account by default
    const [account1, account2] = await ethers.getSigners();

    const DAO = await ethers.getContractFactory('DAO');
    const dao = await DAO.deploy();
    const Work1 = await ethers.getContractFactory('Work1');
    const work1 = await Work1.deploy(dao.target, WORK1_PRICE);

    await dao.initiateValidationPool(account1, 0, { value: 100 });
    await dao.evaluateOutcome(0);

    return {
      dao, work1, account1, account2,
    };
  }

  it('Should deploy', async () => {
    const { dao, work1, account1 } = await loadFixture(deploy);
    expect(dao).to.exist;
    expect(work1).to.exist;
    expect(await dao.memberCount()).to.equal(1);
    expect(await dao.balanceOf(account1)).to.equal(100);
    expect(await dao.totalSupply()).to.equal(100);
    expect(await work1.stakeCount()).to.equal(0);
  });

  it('Should be able to receive availability stake', async () => {
    const { dao, work1, account1 } = await loadFixture(deploy);
    await dao.stakeAvailability(work1.target, STAKE_AMOUNT, STAKE_DURATION);
    expect(await dao.balanceOf(account1)).to.equal(STAKE_AMOUNT);
    expect(await dao.balanceOf(work1.target)).to.equal(STAKE_AMOUNT);
    expect(await work1.stakeCount()).to.equal(1);
    const stake = await work1.stakes(0);
    expect(stake.worker).to.equal(account1);
    expect(stake.amount).to.equal(STAKE_AMOUNT);
    expect(stake.endTime).to.equal(await time.latest() + STAKE_DURATION);
  });

  it('should be able to request work and assign to a worker', async () => {
    const {
      dao, work1, account1, account2,
    } = await loadFixture(deploy);
    await dao.stakeAvailability(work1.target, STAKE_AMOUNT, STAKE_DURATION, { from: account1 });
    const requestWork = () => work1.connect(account2).requestWork({ value: WORK1_PRICE });
    await expect(requestWork()).to.emit(work1, 'WorkAssigned').withArgs(account1, 0);
    expect(await work1.requestCount()).to.equal(1);
    const request = await work1.requests(0);
    expect(request.customer).to.equal(account2);
  });

  it('should not be able to request work if there are no availability stakes', async () => {
    const {
      work1, account2,
    } = await loadFixture(deploy);
    const requestWork = () => work1.connect(account2).requestWork({ value: WORK1_PRICE });
    await expect(requestWork()).to.be.revertedWith('No available worker stakes');
  });

  it('should not assign work to an expired availability stake', async () => {
    const {
      dao, work1, account1, account2,
    } = await loadFixture(deploy);
    await dao.stakeAvailability(work1.target, STAKE_AMOUNT, STAKE_DURATION, { from: account1 });
    const requestWork = () => work1.connect(account2).requestWork({ value: WORK1_PRICE });
    await time.increase(61);
    await expect(requestWork()).to.be.revertedWith('No available worker stakes');
  });
  it('should not assign work to the same availability stake twice', async () => {
    const {
      dao, work1, account1, account2,
    } = await loadFixture(deploy);
    await dao.stakeAvailability(work1.target, STAKE_AMOUNT, STAKE_DURATION, { from: account1 });
    const requestWork = () => work1.connect(account2).requestWork({ value: WORK1_PRICE });
    await expect(requestWork()).to.emit(work1, 'WorkAssigned').withArgs(account1, 0);
    await expect(requestWork()).to.be.revertedWith('No available worker stakes');
  });

  it('should be able to submit work evidence', async () => {
    const {
      dao, work1, account1, account2,
    } = await loadFixture(deploy);
    await dao.stakeAvailability(work1.target, STAKE_AMOUNT, STAKE_DURATION, { from: account1 });
    await work1.connect(account2).requestWork({ value: WORK1_PRICE });
    await expect(work1.submitWorkEvidence(0)).to.emit(work1, 'WorkEvidenceSubmitted').withArgs(0);
  });

  it('should be able to submit work approval', async () => {
    const {
      dao, work1, account1, account2,
    } = await loadFixture(deploy);
    await dao.stakeAvailability(work1.target, STAKE_AMOUNT, STAKE_DURATION, { from: account1 });
    await work1.connect(account2).requestWork({ value: WORK1_PRICE });
    await work1.submitWorkEvidence(0);
    await expect(work1.submitWorkApproval(0, true)).to.emit(dao, 'ValidationPoolInitiated').withArgs(1);
  });
});
